library ieee;
use ieee.std_logic_1164.all;

entity AllButFSM is
	port (
		DataOutEnable  : in std_logic;
		DataInEnable   : in std_logic;
		RAEnable       : in std_logic;
		RBEnable       : in std_logic;
		RCEnable       : in std_logic;
		SPEnable       : in std_logic;
		PCEnable       : in std_logic;
		AddrEnable     : in std_logic;
		IREnable       : in std_logic;
		Mux1Sel        : in std_logic_vector(1 downto 0);
		Mux2Sel        : in std_logic_vector(1 downto 0);
		Mux3Sel        : in std_logic_vector(2 downto 0);
		Mux4Sel        : in std_logic_vector(2 downto 0);
		Mux5Sel        : in std_logic;
		Mux6Sel        : in std_logic;
		AluOpSel       : in std_logic_vector(3 downto 0);
		MemWriteEnable : in std_logic;
		SignExpSel     : in std_logic;
		IRValue        : out std_logic_vector(15 downto 0);
		GPOut          : out std_logic_vector(15 downto 0);
		Clk            : in std_logic;
		Reset          : in std_logic
	);
end entity;

architecture Architecture1 of AllButFSM is
	component Reg is
		generic (
			NBits : integer := 16
		);
		port (
			Enable  : in std_logic;
			Clk     : in std_logic;
			DataIn  : in std_logic_vector((NBits - 1) downto 0);
			DataOut : out std_logic_vector((NBits - 1) downto 0)
		);
	end component;
	component Mux2Inputs is
		generic (
			NBits : integer := 16
		);
		port (
			Sel     : in std_logic;
			DataIn0 : in std_logic_vector((NBits - 1) downto 0);
			DataIn1 : in std_logic_vector((NBits - 1) downto 0);
			DataOut : out std_logic_vector((NBits - 1) downto 0)
		);
	end component;
	component Mux3Inputs is
		generic (
			NBits : integer := 16
		);
		port (
			Sel     : in std_logic_vector(1 downto 0);
			DataIn0 : in std_logic_vector((NBits - 1) downto 0);
			DataIn1 : in std_logic_vector((NBits - 1) downto 0);
			DataIn2 : in std_logic_vector((NBits - 1) downto 0);
			DataOut : out std_logic_vector((NBits - 1) downto 0)
		);
	end component;
	component Mux5Inputs is
		generic (
			NBits : integer := 16
		);
		port (
			Sel     : in std_logic_vector(2 downto 0);
			DataIn0 : in std_logic_vector((NBits - 1) downto 0);
			DataIn1 : in std_logic_vector((NBits - 1) downto 0);
			DataIn2 : in std_logic_vector((NBits - 1) downto 0);
			DataIn3 : in std_logic_vector((NBits - 1) downto 0);
			DataIn4 : in std_logic_vector((NBits - 1) downto 0);
			DataOut : out std_logic_vector((NBits - 1) downto 0)
		);
	end component;
	component Mux6Inputs is
		generic (
			NBits : integer := 16
		);
		port (
			Sel     : in std_logic_vector(2 downto 0);
			DataIn0 : in std_logic_vector((NBits - 1) downto 0);
			DataIn1 : in std_logic_vector((NBits - 1) downto 0);
			DataIn2 : in std_logic_vector((NBits - 1) downto 0);
			DataIn3 : in std_logic_vector((NBits - 1) downto 0);
			DataIn4 : in std_logic_vector((NBits - 1) downto 0);
			DataIn5 : in std_logic_vector((NBits - 1) downto 0);
			DataOut : out std_logic_vector((NBits - 1) downto 0)
		);
	end component;
	component SignExp is
		generic (
			NBitsToExpand0 : integer := 15;
			NBitsToExpand1 : integer := 12;
			NBits : integer := 16
		);
		port (
			SelIn   : in std_logic;
			DataIn  : in std_logic_vector((NBits - 1) downto 0);
			DataOut : out std_logic_vector((NBits - 1) downto 0)
		);
	end component;
	component Alu is
		generic (
			NBits : integer := 16
		);
		port (
			DIn   : in std_logic_vector((NBits - 1) downto 0);
			SIn   : in std_logic_vector((NBits - 1) downto 0);
			RAIn  : in std_logic_vector((NBits - 1) downto 0);
			DOut  : out std_logic_vector((NBits - 1) downto 0);
			SelIn : in std_logic_vector(3 downto 0)
		);
	end component;
	component Memory is
		port (
			AddressIn   : in std_logic_vector(15 downto 0);
			DataIn      : in std_logic_vector(15 downto 0);
			DataOut     : out std_logic_vector(15 downto 0);
			WriteEnable : in std_logic;
			GPOut       : out std_logic_vector(15 downto 0);
			Clk         : in std_logic
		);
	end component;
	signal RAOut      : std_logic_vector(15 downto 0);
	signal DataInOut  : std_logic_vector(15 downto 0);
	signal Mux1Out    : std_logic_vector(15 downto 0);
	signal IROut      : std_logic_vector(15 downto 0);
	signal Mux2Out    : std_logic_vector(15 downto 0);
	signal RBOut      : std_logic_vector(15 downto 0);
	signal RCOut      : std_logic_vector(15 downto 0);
	signal SPOut      : std_logic_vector(15 downto 0);
	signal PCOut      : std_logic_vector(15 downto 0);
	signal SignExpOut : std_logic_vector(15 downto 0);
	signal Mux3Out    : std_logic_vector(15 downto 0);
	signal Mux4Out    : std_logic_vector(15 downto 0);
	signal Mux5Out    : std_logic_vector(3 downto 0);
	signal Mux6Out    : std_logic_vector(15 downto 0);
	signal AluOut     : std_logic_vector(15 downto 0);
	signal AddrOut    : std_logic_vector(15 downto 0);
	signal MemOut     : std_logic_vector(15 downto 0);
	signal MemIn      : std_logic_vector(15 downto 0);
begin
	-- data in and data out registers
	DataOut : Reg generic map (
		NBits => 16
	)
	port map (
		Enable => DataOutEnable,
		Clk => Clk,
		DataIn => RAOut,
		DataOut => MemIn    -- to write RAM
	);
	DataIn : Reg generic map (
		NBits => 16
	)
	port map (
		Enable => DataInEnable,
		Clk => Clk,
		DataIn => MemOut,   -- from read ROM/RAM
		DataOut => DataInOut
	);
	-- addr and ir registers
	Addr : Reg generic map (
		NBits => 16
	)
	port map (
		Enable => AddrEnable,
		Clk => Clk,
		DataIn => Mux1Out,
		DataOut => AddrOut    -- to write ROM/RAM address
	);
	IR : Reg generic map (
		NBits => 16
	)
	port map (
		Enable => IREnable,
		Clk => Clk,
		DataIn => DataInOut,
		DataOut => IROut
	);
	IRValue <= IROut;
	-- programming registers (RA, RB, RC, SP and PC)
	RA : Reg generic map (
		NBits => 16
	)
	port map (
		Enable => RAEnable,
		Clk => Clk,
		DataIn => Mux2Out,
		DataOut => RAOut
	);
	RB : Reg generic map (
		NBits => 16
	)
	port map (
		Enable => RBEnable,
		Clk => Clk,
		DataIn => AluOut,
		DataOut => RBOut
	);
	RC : Reg generic map (
		NBits => 16
	)
	port map (
		Enable => RCEnable,
		Clk => Clk,
		DataIn => AluOut,
		DataOut => RCOut
	);
	SP : Reg generic map (
		NBits => 16
	)
	port map (
		Enable => SPEnable,
		Clk => Clk,
		DataIn => AluOut,
		DataOut => SPOut
	);
	PC : Reg generic map (
		NBits => 16
	)
	port map (
		Enable => PCEnable,
		Clk => Clk,
		DataIn => Mux6Out,
		DataOut => PCOut
	);
	-- muxes
	Mux1 : Mux3Inputs generic map (
		NBits => 16
	)
	port map (
		Sel => Mux1Sel,
		DataIn0 => RBOut,
		DataIn1 => SPOut,
		DataIn2 => PCOut,
		DataOut => Mux1Out
	);
	Mux2 : Mux3Inputs generic map (
		NBits => 16
	)
	port map (
		Sel => Mux2Sel,
		DataIn0 => DataInOut,
		DataIn1 => AluOut,
		DataIn2 => SignExpOut,
		DataOut => Mux2Out
	);
	Mux3 : Mux5Inputs generic map (
		NBits => 16
	)
	port map (
		Sel => Mux3Sel,
		DataIn0 => RAOut,
		DataIn1 => RBOut,
		DataIn2 => RCOut,
		DataIn3 => SPOut,
		DataIn4 => PCOut,
		DataOut => Mux3Out
	);
	Mux4 : Mux6Inputs generic map (
		NBits => 16
	)
	port map (
		Sel => Mux4Sel,
		DataIn0 => RAOut,
		DataIn1 => RBOut,
		DataIn2 => RCOut,
		DataIn3 => SPOut,
		DataIn4 => PCOut,
		DataIn5 => SignExpOut,
		DataOut => Mux4Out
	);
	Mux5 : Mux2Inputs generic map (
		NBits => 4          -- mux for select operation on alu, not for data
	)
	port map (
		Sel => Mux5Sel,
		DataIn0 => IROut(3 downto 0),
		DataIn1 => AluOpSel,
		DataOut => Mux5Out
	);
	Mux6 : Mux2Inputs generic map (
		NBits => 16
	)
	port map (
		Sel => Mux6Sel,
		DataIn0 => "0000000000000000",     -- reset vector
		DataIn1 => AluOut,
		DataOut => Mux6Out
	);
	-- sign expander
	SignExpander : SignExp generic map (
		NBitsToExpand0 => 15,
		NBitsToExpand1 => 12,
		NBits => 16
	)
	port map (
		SelIn => SignExpSel,
		DataIn => IROut,
		DataOut => SignExpOut
	);
	-- alu
	Alu1 : Alu generic map (
		NBits => 16
	)
	port map (
		DIn => Mux3Out,
		SIn => Mux4Out,
		DOut => AluOut,
		RAIn => RAOut,
		SelIn => Mux5Out
	);
	-- memory
	Mem : Memory port map (
		AddressIn => AddrOut,
		DataIn => MemIn,
		DataOut => MemOut,
		WriteEnable => MemWriteEnable,
		GPOut => GPOut,
		Clk => Clk
	);
end architecture;
