library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity Alu is
	generic (
		NBits : integer := 16
	);
	port (
		DIn   : in std_logic_vector((NBits - 1) downto 0);
		SIn   : in std_logic_vector((NBits - 1) downto 0);
		RAIn  : in std_logic_vector((NBits - 1) downto 0);
		DOut  : out std_logic_vector((NBits - 1) downto 0);
		SelIn : in std_logic_vector(3 downto 0)
	);
end entity;

architecture Architecture1 of Alu is
	component Adder is
		generic (
			NBits : integer := 16
		);
		port (
			A : in std_logic_vector((NBits - 1) downto 0);
			B : in std_logic_vector((NBits - 1) downto 0);
			Y : out std_logic_vector((NBits - 1) downto 0)
		);
	end component;
	component Neg is
		generic (
			NBits : integer := 16
		);
		port(
			DataIn  : in std_logic_vector((NBits - 1) downto 0);
			DataOut : out std_logic_vector((NBits - 1) downto 0)
		);
	end component;
	signal AddOut : std_logic_vector((NBits - 1) downto 0);
	signal SInNeg : std_logic_vector((NBits - 1) downto 0);
	signal FirstOperand : std_logic_vector((NBits - 1) downto 0);
	signal SMuxOut : std_logic_vector((NBits - 1) downto 0);
begin
	-- mux for the first operand of the adder
	FirstOperand <= std_logic_vector(to_signed(1, NBits)) when (SelIn = "0111") else                              -- inc
	                std_logic_vector(to_signed(-1, NBits)) when (SelIn = "1000") else                             -- dec
	                DIn when ((SelIn = "0001") or (SelIn = "0010") or (SelIn = "1100") or (SelIn = "1101")) else  -- add, sub, jz, jn
	                (others => '0');                                                                              -- assign
	-- neg Y
	Negate : Neg generic map (
		NBits => NBits
	)
	port map (
		DataIn => SIn,
		DataOut => SInNeg
	);
	-- src mux
	SMuxOut <= SInNeg when (SelIn = "0010") else  -- sub
	           (others => '0') when ((SelIn = "1100") and (RAIn /= std_logic_vector(to_unsigned(0, NBits)))) or ((SelIn = "1101") and (RaIn(NBits - 1) = '0')) else  -- jz, jn
	           SIn;
	-- adder
	Add : Adder generic map (
		NBits => NBits
	)
	port map (
		A => FirstOperand,
		B => SMuxOut,
		Y => AddOut
	);
	-- final mux
	DOut <= AddOut when ((SelIn = "0000") or (SelIn = "0001") or (SelIn = "0010") or (SelIn = "0111") or (SelIn = "1000") or (SelIn = "1100") or (SelIn = "1101")) else
	        (DIn and SIn) when (SelIn = "0011") else
	        (DIn or SIn) when (SelIn = "0100") else
	        (DIn xor SIn) when (SelIn = "0101") else
	        (not(SIn)) when (SelIn = "0110") else
			('0' & SIn(15 downto 1)) when (SelIn = "1001") else         -- slr
			(SIn(15) & SIn(15 downto 1)) when (SelIn = "1010") else     -- sar
			(SIn(14 downto 0) & '0') when (SelIn = "1011") else         -- sll
			(others => '0');
end architecture;
