library ieee;
use ieee.std_logic_1164.all;

entity FSM is
	port (
		DataOutEnable  : out std_logic;
		DataInEnable   : out std_logic;
		RAEnable       : out std_logic;
		RBEnable       : out std_logic;
		RCEnable       : out std_logic;
		SPEnable       : out std_logic;
		PCEnable       : out std_logic;
		AddrEnable     : out std_logic;
		IREnable       : out std_logic;
		Mux1Sel        : out std_logic_vector(1 downto 0);
		Mux2Sel        : out std_logic_vector(1 downto 0);
		Mux3Sel        : out std_logic_vector(2 downto 0);
		Mux4Sel        : out std_logic_vector(2 downto 0);
		Mux5Sel        : out std_logic;
		Mux6Sel        : out std_logic;
		AluOpSel       : out std_logic_vector(3 downto 0);
		MemWriteEnable : out std_logic;
		SignExpSel     : out std_logic;
		IRValue        : in std_logic_vector(15 downto 0);
		Clk            : in std_logic;
		Reset          : in std_logic
	);
end entity;

architecture Architecture1 of FSM is
	signal QBus : std_logic_vector(4 downto 0);
	signal DBus : std_logic_vector(4 downto 0);
begin

	process (Clk, Reset)
	begin
		if (Clk'event and (Clk = '1')) then
			if (Reset = '1') then
				QBus <= (others => '0');
			else
				QBus <= DBus;
			end if;
		end if;
	end process;

	-- next state logic
	-- for state "00000"   MUX6 := "0", Enable PC
	DBus <= "00001" when ((QBus = "00000") or (QBus = "00101") or (QBus = "01000") or (QBus = "01001") or (QBus = "01011") or (QBus = "01110") or (QBus = "10010") or (QBus = "10011") or (QBus = "10100") or (QBus = "10101")) else                -- MUX1 := PC, Enable ADDR
	        "10110" when (QBus = "00001") else        -- memory wait state
	        "00010" when (QBus = "10110") else        -- Enable DATAIN
	        "00011" when (QBus = "00010") else        -- Enable IR
	        "00100" when (QBus = "00011") else        -- MUX5 := FSM, ALU := inc, MUX4 := PC, Enable PC
	        -- LOADI states
	        "00101" when ((QBus = "00100") and (IRValue(15) = '1')) else                  -- MUX2 := EXP, SignSel := 15 bits, Enable RA
	        -- LOAD states
	        "00110" when ((QBus = "00100") and (IRValue(15 downto 12) = "0001")) else     -- MUX1 := RB, Enable ADDR
	        "10111" when (QBus = "00110") else                                            -- memory wait state
	        "00111" when (QBus = "10111") else                                            -- Enable DATAIN
	        "01000" when (QBus = "00111") else                                            -- MUX2 := DATAIN, Enable RA
	        -- OP states
	        "01001" when ((QBus = "00100") and (IRValue(15 downto 12) = "0010")) else     -- MUX2 := ALU, MUX3 := dst, MUX4 := src, MUX5 := IR(3..0), Enable dst
	        -- STORE states
	        "01010" when ((QBus = "00100") and (IRValue(15 downto 12) = "0011")) else     -- MUX1 := RB, Enable ADDR, Enable DATAOUT
	        "01011" when (QBus = "01010") else                                            -- WE := 1
	        -- PUSH states
	        "01100" when ((QBus = "00100") and (IRValue(15 downto 12) = "0100")) else     -- MUX4 := SP, MUX5 := FSM, ALU := inc, Enable SP
	        "01101" when (QBus = "01100") else                                            -- MUX1 := SP, Enable ADDR, Enable DATAOUT
	        "01110" when (QBus = "01101") else                                            -- WE := 1
	        -- POP states
	        "01111" when ((QBus = "00100") and (IRValue(15 downto 12) = "0101")) else     -- MUX1 := SP, Enable ADDR
	        "11000" when (QBus = "01111") else                                            -- memory wait state
	        "10000" when (QBus = "11000") else                                            -- Enable DATAIN
	        "10001" when (QBus = "10000") else                                            -- Enable RA, MUX2 := DATAIN
	        "10010" when (QBus = "10001") else                                            -- MUX4 := SP, MUX5 := FSM, ALU := dec, Enable SP
	        -- J states
	        "10011" when ((QBus = "00100") and (IRValue(15 downto 12) = "0110")) else     -- MUX4 := EXP, SignSel := 12 bits, MUX3 := PC, MUX5 := FSM, ALU := add, Enable PC
	        -- JZ states
	        "10100" when ((QBus = "00100") and (IRValue(15 downto 12) = "0111")) else     -- MUX4 := EXP, SignSel := 12 bits, MUX3 := PC, MUX5 := FSM, ALU := add if RA=0, Enable PC
	        -- JN states
	        "10101" when ((QBus = "00100") and (IRValue(15 downto 12) = "0000")) else     -- MUX4 := EXP, SignSel := 12 bits, MUX3 := PC, MUX5 := FSM, ALU := add if RA<0, Enable PC
			"00000";

	-- output logic
	DataOutEnable <= '1' when (QBus = "01010") or (QBus = "01101") else
	                 '0';
	DataInEnable <= '1' when (QBus = "00010") or (QBus = "00111") or (QBus = "10000") else
	                '0';
	RAEnable <= '1' when (QBus = "00101") or (QBus = "01000") or (QBus = "10001") or ((QBus = "01001") and (IRValue(10 downto 8) = "000")) else
	            '0';
	RBEnable <= '1' when (QBus = "01001") and (IRValue(10 downto 8) = "001") else
	            '0';
	RCEnable <= '1' when (QBus = "01001") and (IRValue(10 downto 8) = "010") else
	            '0';
	SPEnable <= '1' when (QBus = "01100") or (QBus = "10010") or ((QBus = "01001") and (IRValue(10 downto 8) = "011")) else
	            '0';
	PCEnable <= '1' when (QBus = "00000") or (QBus = "00100") or (QBus = "10011") or (QBus = "10100") or (QBus = "10101") or ((QBus = "01001") and (IRValue(10 downto 8) = "100")) else
	            '0';
	AddrEnable <= '1' when (QBus = "00001") or (QBus = "00110") or (QBus = "01010") or (QBus = "01101") or (QBus = "01111") else
	              '0';
	IREnable <= '1' when (QBus = "00011") else
	            '0';
	Mux1Sel <= "00" when (QBus = "00110") or (QBus = "01010") else
	           "01" when (QBus = "01101") or (QBus = "01111") else
	           "10";
	Mux2Sel <= "00" when (QBus = "01000") or (QBus = "10001") else
	           "01" when (QBus = "01001") else
	           "10";
	Mux3Sel <= "100" when (QBus = "10011") or (QBus = "10100") or (QBus = "10101") else
	           IRValue(10 downto 8);
	Mux4Sel <= "011" when (QBus = "01100") or (QBus = "10010") else
	           "100" when (QBus = "00100") else
	           "101" when (QBus = "10011") or (QBus = "10100") or (QBus = "10101") else
	           IRValue(6 downto 4);
	Mux5Sel <= '0' when (QBus = "01001") else
	           '1';
	Mux6Sel <= '0' when (QBus = "00000") else
	           '1';
	AluOpSel <= "0111" when (QBus = "00100") or (QBus = "01100") else
	            "1000" when (QBus = "10010") else
	            "0001" when (QBus = "10011") else
	            "1100" when (QBus = "10100") else
	            "1101" when (QBus = "10101") else
	            "0000";
	MemWriteEnable <= '1' when (QBus = "01011") or (QBus = "01110") else
	                  '0';
	SignExpSel <= '0' when (QBus = "00101") else
	              '1';
end architecture;
