library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity Memory is
	generic (
		GPOutAddress : integer := 12288
	);
	port (
		AddressIn   : in std_logic_vector(15 downto 0);
		DataIn      : in std_logic_vector(15 downto 0);
		DataOut     : out std_logic_vector(15 downto 0);
		WriteEnable : in std_logic;
		GPOut       : out std_logic_vector(15 downto 0);
		Clk         : in std_logic
	);
end entity;

architecture Architecture1 of Memory is
	component Rom is
		port (
			AddressIn : in std_logic_vector(12 downto 0);
			DataOut   : out std_logic_vector(15 downto 0)
		);
	end component;
	component Ram is
		port (
			AddressIn   : in std_logic_vector(12 downto 0);
			DataOut     : out std_logic_vector(15 downto 0);
			DataIn      : in std_logic_vector(15 downto 0);
			WriteEnable : in std_logic;
			Enable      : in std_logic;
			Clk         : in std_logic
		);
	end component;
	signal RomOut : std_logic_vector(15 downto 0);
	signal RamOut : std_logic_vector(15 downto 0);
	signal GPOutD : std_logic_vector(15 downto 0);
	signal GPOutQ : std_logic_vector(15 downto 0);
begin
	-- GPOut register
	process (Clk)
	begin
		if ((Clk'event) and (Clk = '1')) then
			GPOutQ <= GPOutD;
		end if;
	end process;
	GPOutD <= DataIn when (to_integer(unsigned(AddressIn)) = GPOutAddress) else
	          GPOutQ;
	GPOut <= GPOutQ;

	-- ROM
	Rom1 : Rom port map (    -- 8192 words (16 bit)
		AddressIn => AddressIn(12 downto 0),
		DataOut => RomOut
	);

	-- RAM
	Ram1 : Ram port map (    -- 8192 words (16 bit)
		AddressIn => AddressIn(12 downto 0),
		DataOut => RamOut,
		WriteEnable => WriteEnable,
		Enable => '1',
		DataIn => DataIn,
		Clk => Clk
	);

	-- RAM vs ROM multiplexer
	DataOut <= RomOut when (AddressIn(13) = '0') else     -- bit 13 = 0 --> ROM,  bit 13 = 1 --> RAM
	           RamOut;
end architecture;
